import numpy as np
import pandas as pd
from sklearn.model_selection import GroupShuffleSplit as sklearnGroupShuffleSplit

def test(df):
    df["new_col"] = np.nan
    return df


def GroupShuffleSplit(df_main, train_pct=.6, test_pct=.25, val_pct=.25, group_col = 'img_filename', random_state=None):
    """
    This function uses the GroupShuffleSplit command from sklearn. It can split into 3 groups (train,
    test, and val) by applying the command twice. 
    """

    gss = sklearnGroupShuffleSplit(n_splits=1, train_size=train_pct)
    train_indexes, test_indexes = next(gss.split(X=df_main, y=df_main[group_col], groups=df_main.index.values))

    df_main.loc[train_indexes,'split'] = "train"
    df_main.loc[test_indexes,'split'] = "test"

    if val_pct:
        df_train = df_main.loc[df_main['split'] == 'train']
        df_test = df_main.loc[df_main['split'] == 'test']
        df_test = df_test.reset_index()
        second_split_pct = float(test_pct/(test_pct+val_pct))
        gss2 = sklearnGroupShuffleSplit(n_splits=1, train_size=second_split_pct)
        test_indexes_2, val_indexes_2 = next(gss2.split(X=df_test, y=df_test[group_col], groups=df_test.index.values))
        df_test.loc[test_indexes_2,'split'] = "test"
        df_test.loc[val_indexes_2,'split'] = "val"
        return df_train.append(df_test)

    else:
        return df_main


#Written with the help of https://stackoverflow.com/questions/56872664/complex-dataset-split-stratifiedgroupshufflesplit 
def StratifiedGroupShuffleSplit(df_main, train_pct=.7, test_pct=.3, val_pct=.0, weight=0.01, 
    group_col = 'img_filename', cat_col = 'cat_name', batch_size=1):
    """
    This function will 'split" the dataframe by setting the split collumn equal to 
    train, test, or val. When a split dataset is exported the annotations will be split into
    seperate groups so that can be used used in model training, testing, and validation.
    """

    df_main = df_main.reindex(np.random.permutation(df_main.index)) # shuffle dataset

    # create empty train, val and test datasets
    df_train = pd.DataFrame()
    df_val = pd.DataFrame()
    df_test = pd.DataFrame()

    subject_grouped_df_main = df_main.groupby([group_col], sort=False, as_index=False)
    category_grouped_df_main = df_main.groupby(cat_col).count()[[group_col]]/len(df_main)*100

    #Check inputs 
    assert (0 <= weight <= 1), "Weight must be between 0 and 1"
    total_splits = round((train_pct) + float(test_pct) + float(val_pct),1)
    assert (total_splits == 1), "Sum of train_pct, test_pct, and val_pct must equal 1."
    assert (batch_size >= 1 and batch_size <= subject_grouped_df_main.ngroups / 10 ), \
        "Batch must be greater than 1 and less than 1/10 count of groups"

    def calc_mse_loss(df):
        grouped_df = df.groupby(cat_col).count()[[group_col]]/len(df)*100
        df_temp = category_grouped_df_main.join(grouped_df, on = cat_col, how = 'left', lsuffix = '_main')
        df_temp.fillna(0, inplace=True)
        df_temp['diff'] = (df_temp['img_filename_main'] - df_temp[group_col])**2
        mse_loss = np.mean(df_temp['diff'])
        return mse_loss

    i = 0 #counter for all items in dataset
    b = 0 #counter for the batches
    batch_df = df_main[0:0]

    for _, group in subject_grouped_df_main:
        if (i < 3):
            if (i == 0):
                df_train = df_train.append(pd.DataFrame(group), ignore_index=True)
                i += 1
                continue
            elif (i == 1):
                df_val = df_val.append(pd.DataFrame(group), ignore_index=True)
                i += 1
                continue
            else:
                df_test = df_test.append(pd.DataFrame(group), ignore_index=True)
                i += 1
                continue

        #Add groups to the 
        batch_df = batch_df.append(group)
        b += 1
        if b < batch_size and i < subject_grouped_df_main.ngroups-3:
            i += 1
            continue

        mse_loss_diff_train = calc_mse_loss(df_train) - calc_mse_loss(df_train.append(batch_df, ignore_index=True))
        mse_loss_diff_val = calc_mse_loss(df_val) - calc_mse_loss(df_val.append(batch_df, ignore_index=True))
        mse_loss_diff_test = calc_mse_loss(df_test) - calc_mse_loss(df_test.append(batch_df, ignore_index=True))

        total_records = len(df_train) + len(df_val) + len(df_test)

        len_diff_train = (train_pct - (len(df_train)/total_records))
        len_diff_val = (val_pct - (len(df_val)/total_records))
        len_diff_test = (test_pct - (len(df_test)/total_records)) 

        len_loss_diff_train = len_diff_train * abs(len_diff_train)
        len_loss_diff_val = len_diff_val * abs(len_diff_val)
        len_loss_diff_test = len_diff_test * abs(len_diff_test)

        loss_train = (weight * mse_loss_diff_train) + ((1-weight) * len_loss_diff_train)
        loss_val = (weight * mse_loss_diff_val) + ((1-weight) * len_loss_diff_val)
        loss_test = (weight * mse_loss_diff_test) + ((1-weight) * len_loss_diff_test)

        if (max(loss_train,loss_val,loss_test) == loss_train):
            df_train = df_train.append(batch_df, ignore_index=True)
        elif (max(loss_train,loss_val,loss_test) == loss_val):
            df_val = df_val.append(batch_df, ignore_index=True)
        else:
            df_test = df_test.append(batch_df, ignore_index=True)

        #print ("Group " + str(i) + ". loss_train: " + str(loss_train) + " | " + "loss_val: " + str(loss_val) + " | " + "loss_test: " + str(loss_test) + " | ")
        i += 1
        #Reset the batch
        b = 0
        batch_df = df_main[0:0]


    ######
    # Final prep tasks before returning the split dataframe

    #Sometimes the algo will put some rows in the val set even if the split percent was set to zero
    #In those cases move the rows from val to test 
    if round(val_pct,1) == round(0,1):
        df_test.append(df_val)
        df_val = df_val[0:0] #remove the values from 

    #Apply train, split, val labels to the split collumn 
    df_train['split'] = 'train'
    df_test['split'] = 'test'
    df_val['split'] = 'val'

    df = df_train.append(df_test).append(df_val)
    
    assert df.shape == df_main.shape, "Output shape does not match input shape. Data loss has occured."

    return df


import json
from typing import List
import pandas as pd
import numpy as np
import xml.etree.ElementTree as ET
import xml.dom.minidom
import os 
from pathlib import PurePath, Path

class Export():
    def  __init__(self, dataset=None):
        self.dataset = dataset
        
    def ExportToVoc(self, output_path=None, segmented_=False, path_=False, database_=False, folder_=False, occluded_=False):
        ds = self.dataset

        if output_path == None:
            output_path = ds.path_to_annotations
        else:        
            output_path = output_path

        os.makedirs(output_path, exist_ok=True)

        """ Writes annotation files to disk in VOC XML format and returns path to files.

        By default, tags with empty values will not be included in the XML output. 
        You can optionally choose to include them if they are required for your solution.

        Args:
            dataset (obj): 
                A dataset object that contains the annotations and metadata.
            output_path (str or None): 
                This is where the annotation files will be written.
                If not-specified then the path will be derived from the .path_to_annotations and
                .name properties of the dataset object. 

        Returns:
            A list with 1 or more paths (strings) to annotations files.
        """   
        output_file_paths = []

        def voc_xml_file_creation(data, file_name, output_file_path, segmented=True, path=True, database=True, folder=True, occluded=True):
         
            index = 0
            df_smaller = data[data['img_filename'] == file_name].reset_index()
            
            if len(df_smaller) == 1:
                #print('test')
                annotation_text_start = '<annotation>'

                flder_lkp = str(df_smaller.loc[index]['img_folder'])
                if folder==True and flder_lkp != '':
                    folder_text = '<folder>'+flder_lkp+'</folder>'
                else:
                    folder_text = ''
                    
                filename_text = '<filename>'+str(df_smaller.loc[index]['img_filename'])+'</filename>'
                
                pth_lkp = str(df_smaller.loc[index]['img_path'])
                if path == True and pth_lkp != '':
                    path_text = '<path>'+ pth_lkp +'</path>'
                else:
                    path_text = ''
                    
                sources_text = ''
                
                size_text_start = '<size>'
                width_text = '<width>'+str(df_smaller.loc[index]['img_width'])+'</width>'
                height_text = '<height>'+str(df_smaller.loc[index]['img_height'])+'</height>'
                depth_text = '<depth>'+str(df_smaller.loc[index]['img_depth'])+'</depth>'
                size_text_end = '</size>'
                
                seg_lkp = str(df_smaller.loc[index]['ann_segmented'])
                if segmented == True and seg_lkp != '':
                    segmented_text = '<segmented>'+str(df_smaller.loc[index]['ann_segmented'])+'</segmented>'
                else:
                    segmented_text = ''

                object_text_start = '<object>'

                name_text = '<name>'+str(df_smaller.loc[index]['cat_name'])+'</name>'
                pose_text = '<pose>'+str(df_smaller.loc[index]['ann_pose'])+'</pose>'
                truncated_text = '<truncated>'+str(df_smaller.loc[index]['ann_truncated'])+'</truncated>'
                difficult_text = '<difficult>'+str(df_smaller.loc[index]['ann_difficult'])+'</difficult>'
                
                occluded_text = ''

                bound_box_text_start = '<bndbox>'

                xmin_text = '<xmin>'+str(df_smaller.loc[index]['ann_bbox_xmin'])+'</xmin>'
                xmax_text = '<xmax>'+str(df_smaller.loc[index]['ann_bbox_xmax'])+'</xmax>'
                ymin_text = '<ymin>'+str(df_smaller.loc[index]['ann_bbox_ymin'])+'</ymin>'
                ymax_text = '<ymax>'+str(df_smaller.loc[index]['ann_bbox_ymax'])+'</ymax>'

                bound_box_text_end = '</bndbox>'
                object_text_end = '</object>'
                annotation_text_end = '</annotation>'
                        
                xmlstring = annotation_text_start + folder_text  +filename_text  + \
                    path_text  + sources_text + size_text_start + width_text  + \
                    height_text  + depth_text  + size_text_end + segmented_text  + \
                    object_text_start + name_text  + pose_text  +truncated_text + \
                    difficult_text + occluded_text + bound_box_text_start  +xmin_text  + \
                    xmax_text  +ymin_text  +ymax_text  +bound_box_text_end  + \
                    object_text_end  + annotation_text_end
                dom = xml.dom.minidom.parseString(xmlstring)
                pretty_xml_as_string = dom.toprettyxml()
                
                with open(output_file_path, "w") as f:
                    f.write(pretty_xml_as_string)    

                return output_file_path
            
            else:

                #print('test')
                annotation_text_start = '<annotation>'
                
                flder_lkp = str(df_smaller.loc[index]['img_folder'])
                if folder==True and flder_lkp != '':
                    folder_text = '<folder>'+flder_lkp+'</folder>'
                else:
                    folder_text = ''
                
                filename_text = '<filename>'+str(df_smaller.loc[index]['img_filename'])+'</filename>'
                
                pth_lkp = str(df_smaller.loc[index]['img_path'])
                if path == True and pth_lkp != '':
                    path_text = '<path>'+ pth_lkp +'</path>'
                else:
                    path_text = ''
                
                #db_lkp = str(df_smaller.loc[index]['Databases'])
                #if database == True and db_lkp != '':
                #    sources_text = '<source>'+'<database>'+ db_lkp +'</database>'+'</source>'
                #else:
                sources_text = ''
                    
                size_text_start = '<size>'
                width_text = '<width>'+str(df_smaller.loc[index]['img_width'])+'</width>'
                height_text = '<height>'+str(df_smaller.loc[index]['img_height'])+'</height>'
                depth_text = '<depth>'+str(df_smaller.loc[index]['img_depth'])+'</depth>'
                size_text_end = '</size>'
                
                seg_lkp = str(df_smaller.loc[index]['ann_segmented'])
                if segmented == True and seg_lkp != '':
                    segmented_text = '<segmented>'+str(df_smaller.loc[index]['ann_segmented'])+'</segmented>'
                else:
                    segmented_text = ''

                xmlstring = annotation_text_start + folder_text  +filename_text  + \
                        path_text  + sources_text + size_text_start + width_text  + \
                        height_text  + depth_text  + size_text_end + segmented_text
                
                for obj in range(len(df_smaller)):
                    object_text_start = '<object>'

                    name_text = '<name>'+str(df_smaller.loc[index]['cat_name'])+'</name>'
                    pose_text = '<pose>'+str(df_smaller.loc[index]['ann_pose'])+'</pose>'
                    truncated_text = '<truncated>'+str(df_smaller.loc[index]['ann_truncated'])+'</truncated>'
                    difficult_text = '<difficult>'+str(df_smaller.loc[index]['ann_difficult'])+'</difficult>'
                    
                    #occ_lkp = str(df_smaller.loc[index]['Object Occluded'])
                    #if occluded==True and occ_lkp != '':
                    #    occluded_text = '<occluded>'+occ_lkp+'</occluded>'
                    #else:
                    occluded_text = ''

                    bound_box_text_start = '<bndbox>'

                    xmin_text = '<xmin>'+str(df_smaller.loc[index]['ann_bbox_xmin'])+'</xmin>'
                    xmax_text = '<xmax>'+str(df_smaller.loc[index]['ann_bbox_xmax'])+'</xmax>'
                    ymin_text = '<ymin>'+str(df_smaller.loc[index]['ann_bbox_ymin'])+'</ymin>'
                    ymax_text = '<ymax>'+str(df_smaller.loc[index]['ann_bbox_ymax'])+'</ymax>'

                    bound_box_text_end = '</bndbox>'
                    object_text_end = '</object>'
                    annotation_text_end = '</annotation>'
                    index = index + 1

                    xmlstring = xmlstring + object_text_start + name_text  + pose_text  +truncated_text + \
                        difficult_text + occluded_text + bound_box_text_start  +xmin_text  + \
                        xmax_text  +ymin_text  +ymax_text  +bound_box_text_end  + \
                        object_text_end  

                xmlstring = xmlstring + annotation_text_end
                dom = xml.dom.minidom.parseString(xmlstring)
                pretty_xml_as_string = dom.toprettyxml()
                
                with open(output_file_path, "w") as f:
                    f.write(pretty_xml_as_string)  

                return output_file_path

        #Loop through all images in the dataframe and call voc_xml_file_creation for each one
        for file_title in list(set(self.dataset.df.img_filename)):

            file_name = Path(file_title)
            file_name = str(file_name.with_suffix('.xml'))
            file_path = str(Path(output_path, file_name))
            voc_file_path = voc_xml_file_creation(ds.df, file_title, segmented=segmented_, path=path_, database=database_, folder=folder_, occluded=occluded_, output_file_path=file_path)
            output_file_paths.append(voc_file_path)

        return output_file_paths

    def ExportToYoloV5(self, output_path=None):
        """ Writes annotation files to disk and returns path to files.
        
        Args:
            dataset (obj): 
                A dataset object that contains the annotations and metadata.
            output_path (str or None): 
                This is where the annotation files will be written.
                If not-specified then the path will be derived from the .path_to_annotations and
                .name properties of the dataset object. 

        Returns:
            A list with 1 or more paths (strings) to annotations files.
        """
        ds = self.dataset
        #Inspired by https://github.com/aws-samples/groundtruth-object-detection/blob/master/create_annot.py 
        unique_images = ds.df["img_filename"].unique()
        output_file_paths = []

        yolo_dataset = ds.df.copy(deep=True)
        yolo_dataset.cat_id = yolo_dataset.cat_id.astype('float').astype('Int32')
        yolo_dataset["center_x_scaled"] = (yolo_dataset["ann_bbox_xmin"] + (yolo_dataset["ann_bbox_width"]*0.5))/yolo_dataset["img_width"]
        yolo_dataset["center_y_scaled"] = (yolo_dataset["ann_bbox_ymin"] + (yolo_dataset["ann_bbox_height"]*0.5))/yolo_dataset["img_height"]
        yolo_dataset["width_scaled"] = yolo_dataset["ann_bbox_width"] / yolo_dataset["img_width"]
        yolo_dataset["height_scaled"] = yolo_dataset["ann_bbox_height"] / yolo_dataset["img_height"]
        yolo_dataset[["cat_id", "center_x_scaled", "center_y_scaled", "width_scaled", "height_scaled"]]

        #Create folders to store annotations
        if output_path == None:
            dest_folder = PurePath(ds.path_to_annotations, yolo_dataset.iloc[0].img_folder)
        else:
            dest_folder = output_path

        os.makedirs(dest_folder, exist_ok=True)

        for img_filename in unique_images:
                df_single_img_annots = yolo_dataset.loc[yolo_dataset.img_filename == img_filename]
                annot_txt_file = img_filename.split(".")[0] + ".txt"
                destination = f"{dest_folder}/{annot_txt_file}"
                df_single_img_annots.to_csv(
                    destination,
                    index=False,
                    header=False,
                    sep=" ",
                    float_format="%.4f",
                    columns=[
                        "cat_id",
                        "center_x_scaled",
                        "center_y_scaled",
                        "width_scaled",
                        "height_scaled",
                    ])
                output_file_paths.append(destination)

        return output_file_paths

    def ExportToCoco(self, output_path=None) -> List:
        """ 
        Writes annotation files to disk and returns path to files.

        Args:
            dataset (obj): 
                A dataset object that contains the annotations and metadata.
            output_path (str or None): 
                This is where the annotation files will be written.
                If not-specified then the path will be derived from the .path_to_annotations and
                .name properties of the dataset object. 

        Returns:
            A list with 1 or more paths (strings) to annotations files.
        """
        df = self.dataset.df
        df_outputI = []
        df_outputA = []
        df_outputC = []
        list_i = []
        list_c = []
        json_list = []
        
        for i in range(0,df.shape[0]):
            images = [{
            "id": df['img_id'][i], 
            "folder": df['img_folder'][i], 
            "file_name": df['img_filename'][i], 
            "path": df['img_path'][i], 
            "width": df['img_width'][i], 
            "height": df['img_height'][i], 
            "depth": df['img_depth'][i]
            }]
        
            annotations = [{
            "image_id": df['img_id'][i], 
            "id": df.index[i], 
            "segmented": df['ann_segmented'][i],
            "bbox": [df['ann_bbox_xmin'][i], df['ann_bbox_ymin'][i], df['ann_bbox_width'][i], df['ann_bbox_height'][i]],  
            "area": df['ann_area'][i], 
            "segmentation": df['ann_segmentation'][i], 
            "iscrowd": df['ann_iscrowd'][i], 
            "pose": df['ann_pose'][i], 
            "truncated": df['ann_truncated'][i],
            "category_id": df['cat_id'][i],  
            "difficult": df['ann_difficult'][i]
            }]

            categories = [{
            "id": int(df['cat_id'][i]), 
            "name": df['cat_name'][i], 
            "supercategory": df['cat_supercategory'][i]
            }]
            
            if(list_c):
                if (categories[0]["id"] in list_c or np.isnan(categories[0]["id"])):
                    pass    
                else:
                    df_outputC.append(pd.DataFrame([categories]))
            elif(~np.isnan(categories[0]["id"])):
                df_outputC.append(pd.DataFrame([categories]))
            else:
                pass
            list_c.append(categories[0]["id"])

            if(list_i):
                if (images[0]["id"] in list_i or np.isnan(images[0]["id"])):
                    pass
                else:
                    df_outputI.append(pd.DataFrame([images]))
            elif(~np.isnan(images[0]["id"])):
                df_outputI.append(pd.DataFrame([images]))      
            else:
                pass
            list_i.append(images[0]["id"])    

            df_outputA.append(pd.DataFrame([annotations]))
            
        mergedI = pd.concat(df_outputI, ignore_index=True)
        mergedA = pd.concat(df_outputA, ignore_index=True)
        mergedC = pd.concat(df_outputC, ignore_index=True)
        
        resultI = mergedI[0].to_json(orient="split")
        resultA = mergedA[0].to_json(orient="split")
        resultC = mergedC[0].to_json(orient="split")

        parsedI = json.loads(resultI)
        del parsedI['index']
        del parsedI['name']
        parsedI['images'] = parsedI['data']
        del parsedI['data']

        parsedA = json.loads(resultA)
        del parsedA['index']
        del parsedA['name']
        parsedA['annotations'] = parsedA['data']
        del parsedA['data']

        parsedC = json.loads(resultC)
        del parsedC['index']
        del parsedC['name']
        parsedC['categories'] = parsedC['data']
        del parsedC['data']

        parsedI.update(parsedA)
        parsedI.update(parsedC)
        json_output = parsedI

        if output_path == None:
            output_path = Path(self.dataset.path_to_annotations, (self.dataset.name + ".json"))
            
        with open(output_path, 'w') as outfile:
            json.dump(obj=json_output, fp=outfile, indent=4)
        return [str(output_path)] 


from pylabel.analyze import Analyze
from pylabel.exporter import Export
from pylabel.visualize import Visualize

from pylabel.splitter import GroupShuffleSplit, test
from pylabel.splitter import StratifiedGroupShuffleSplit


class Dataset:
    def __init__(self, df):
        self.df = df
        self.name = "dataset"
        self.path_to_annotations = ""    
        self.GroupShuffleSplit = GroupShuffleSplit
        self.test = test
        self.StratifiedGroupShuffleSplit = StratifiedGroupShuffleSplit
        self.export = Export(dataset=self)
        self.visualize = Visualize(dataset=self)
        self.analyze = Analyze(self)
        #self.splitter = Splitter(dataset=self)




import pandas as pd

class Analyze:
    def  __init__(self, dataset=None):
        self.dataset = dataset 
        ds = self.dataset

        self.class_counts = ds.df["cat_name"].value_counts(dropna=False)
        self.num_classes = ds.df["cat_name"].nunique()
        self.classes = ds.df["cat_name"].unique()
        self.num_images = ds.df["img_filename"].nunique()
        self.split_counts = ds.df["split"].value_counts(dropna=False)
        self.split_pct = ds.df["split"].value_counts(normalize=True, dropna=False)

    def ShowClassSplits(self, normalize=True):
        ds = self.dataset

        def move_column_inplace(df, col, pos):
            """
            Assists to rearrange columns to a desired order. 
            """
            col = df.pop(col)
            df.insert(pos, col.name, col)

        df_value_counts = pd.DataFrame(ds.df["cat_name"].value_counts(normalize=normalize), columns=["cat_name"])

        df_value_counts.index.name = "cat_name"  
        df_value_counts.columns = ["all"]

        split_df = ds.df.groupby('split')

        if split_df.ngroups == 1:
            return df_value_counts

        for name, group in split_df:
            group_df = pd.DataFrame(group)
            df_split_value_counts = pd.DataFrame(group_df["cat_name"].value_counts(normalize=normalize), columns=["cat_name"])
            df_split_value_counts.index.name = "cat_name"
            df_split_value_counts.columns = [name]
            df_value_counts = pd.merge(df_value_counts, df_split_value_counts, how="left", on=["cat_name"])

        #Move 'train' to the left of the table since that is the usual convention.
        if 'train' in df_value_counts.columns:
            move_column_inplace(df_value_counts, 'train', 1)

        return df_value_counts



#########
# This file has variables that are used by the rest of the package 
#########

#These are the valid columns in the pylabel annotations table.              
schema = ['img_folder','img_filename','img_path','img_id','img_width','img_height','img_depth','ann_segmented','ann_bbox_xmin','ann_bbox_ymin','ann_bbox_xmax','ann_bbox_ymax','ann_bbox_width','ann_bbox_height','ann_area','ann_segmentation','ann_iscrowd','ann_pose','ann_truncated','ann_difficult','cat_id','cat_name','cat_supercategory','split']
#schema = ['id','img_folder','img_filename','img_path','img_id','img_width','img_height','img_depth','ann_segmented','ann_bbox_xmin','ann_bbox_ymin','ann_bbox_xmax','ann_bbox_ymax','ann_bbox_width','ann_bbox_height','ann_area','ann_segmentation','ann_iscrowd','ann_pose','ann_truncated','ann_difficult','cat_id','cat_name','cat_supercategory','split']
